from unittest import TestCase

import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd

from clinbin import classify_blood_pressure, diagnose_metabolic_syndrome
from clinbin.blood import _score_mets_ncep_atp_iii_2005


class TestBloodPressure(TestCase):
    def test_blood_pressure(self):
        """Test all types of blood pressure combinations."""
        index = [
            "alice",
            "bob",
            "charlie",
            "dan",
            "eve",
            "frank",
            "george",
            "harry",
            "irene",
            "john",
        ]
        systolic = pd.Series(
            [80, 120, np.nan, 50, 50, 150, 130, 140, np.nan, 60.0], index=index
        )
        diastolic = pd.Series(
            [60, 60, 80, 80, 100, 60, 100, 80, np.nan, np.nan], index=index
        )
        true_levels = pd.Categorical(
            [
                "normal",
                "elevated",
                np.nan,
                "elevated",
                "hypertension",
                "hypertension",
                "hypertension",
                "hypertension",
                np.nan,
                np.nan,
            ],
            categories=["normal", "elevated", "hypertension"],
            ordered=True,
        )
        bp_levels = classify_blood_pressure(systolic, diastolic)
        self.assertTrue(bp_levels.equals(pd.Series(true_levels, index=index)))


class TestMetabolicSyndrome(TestCase):
    def test_ncpe_atp_iii_2005_criterion(self):
        """Test the metabolic syndrome diagnosis."""
        # Asian woman that scores 3/5, but only when applying ethnicity specific cut-off
        # point.
        alice = {
            # Elevated circumference.
            "waist": 0.8,
            # Normal level, but use medication.
            "triglycerides": 1.6,
            "use_triglyceride_meds": True,
            # Reduced HDL cholesterol.
            "hdl_cholesterol": 1.2,
            "use_hdlc_meds": np.nan,
            # Normal blood pressure.
            "systolic": 120,
            "diastolic": 80,
            "use_blood_pressure_meds": None,
            # Normal glucose.
            "glucose": 5.5,
            "ethnicity": "asian",
            "sex": "female",
        }
        # Asian man with similar values scores 2/5.
        bob = {
            # Normal circumference.
            "waist": 0.8,
            # Normal level, but use medication.
            "triglycerides": 1.6,
            "use_triglyceride_meds": True,
            # Normal HDL cholesterol.
            "hdl_cholesterol": 1.2,
            "use_hdlc_meds": np.nan,
            # Normal blood pressure.
            "systolic": 120,
            "diastolic": 80,
            "use_blood_pressure_meds": None,
            # Normal glucose.
            "glucose": 5.5,
            "ethnicity": "asian",
            "sex": "male",
        }
        # White woman scoring 4/5.
        charlie = {
            # Elevated circumference.
            "waist": 0.88,
            # Normal level.
            "triglycerides": 1.6,
            "use_triglyceride_meds": np.nan,
            # Reduced HDL cholesterol.
            "hdl_cholesterol": 1.1,
            "use_hdlc_meds": np.nan,
            # Elevated blood pressure.
            "systolic": 132,
            "diastolic": 85,
            "use_blood_pressure_meds": None,
            # Elevated glucose.
            "glucose": 5.65,
            "ethnicity": "white",
            "sex": "female",
        }
        # White man with similar values scoring 2/5.
        dan = {
            # Normal circumference.
            "waist": 0.88,
            # Normal level.
            "triglycerides": 1.6,
            "use_triglyceride_meds": np.nan,
            # Normal HDL cholesterol.
            "hdl_cholesterol": 1.1,
            "use_hdlc_meds": np.nan,
            # Elevated blood pressure.
            "systolic": 132,
            "diastolic": 84,
            "use_blood_pressure_meds": None,
            # Elevated glucose.
            "glucose": 5.65,
            "ethnicity": "white",
            "sex": "male",
        }
        # Eve does not have all the information to compute all criteria, but has enough
        # to score 4/5.
        eve = {
            # Elevated circumference.
            "waist": 0.88,
            # Missing triglyceride level.
            "triglycerides": np.nan,
            "use_triglyceride_meds": np.nan,
            # Elevated HDL cholesterol.
            "hdl_cholesterol": 1.1,
            "use_hdlc_meds": np.nan,
            # Elevated blood pressure.
            "systolic": 122,
            "diastolic": 85,
            "use_blood_pressure_meds": None,
            # Elevated glucose.
            "glucose": 5.65,
            "ethnicity": "asian",
            "sex": "female",
        }
        # Frank also doesn't have all the information to compute all criteria. We only
        # know that he meets 1≥ criterion. But we don't know enough to meet the 3/5
        # threshold. Thus, we don't know if Frank is positive, so the score is NaN.
        frank = {
            # Normal circumference.
            "waist": 0.88,
            # Missing triglyceride level.
            "triglycerides": np.nan,
            "use_triglyceride_meds": np.nan,
            # Normal HDL cholesterol.
            "hdl_cholesterol": 1.1,
            "use_hdlc_meds": np.nan,
            # Partially missing blood pressure.
            "systolic": np.nan,
            "diastolic": 84,
            "use_blood_pressure_meds": None,
            # Elevated glucose.
            "glucose": 5.65,
            "ethnicity": "asian",
            "sex": "male",
        }
        names = ["alice", "bob", "charlie", "dan", "eve", "frank"]
        data = pd.DataFrame.from_records(
            [alice, bob, charlie, dan, eve, frank], index=names
        )
        score_card = _score_mets_ncep_atp_iii_2005(
            data["waist"],
            data["triglycerides"],
            data["hdl_cholesterol"],
            data["systolic"],
            data["diastolic"],
            data["glucose"],
            data["sex"],
            data["ethnicity"],
            data["use_triglyceride_meds"],
            data["use_hdlc_meds"],
            data["use_blood_pressure_meds"],
        )

        # (1) Waist circumference
        expected_waist_scores = [1, 0, 1, 0, 1, 0]
        assert_array_equal(score_card["waist"], expected_waist_scores)

        # (2) Triglycerides
        expected_trig_scores = [1, 1, 0, 0, np.nan, np.nan]
        assert_array_equal(score_card["triglycerides"], expected_trig_scores)

        # (3) HDL cholesterol
        expected_hdl_scores = [1, 0, 1, 0, 1, 0]
        assert_array_equal(score_card["hdl_cholesterol"], expected_hdl_scores)

        # (4) Blood pressure
        expected_bp_scores = [0, 0, 1, 1, 1, np.nan]
        assert_array_equal(score_card["blood_pressure"], expected_bp_scores)

        # (5) Fasting glucose
        expected_glucose_scores = [0, 0, 1, 1, 1, 1]
        assert_array_equal(score_card["glucose"], expected_glucose_scores)

        # Test diagnosis.
        diagnoses = diagnose_metabolic_syndrome(**data)
        expected_diagnoses = pd.Series(
            [1, 0, 1, 0, 1, pd.NA], index=names, dtype="Int64"
        )
        self.assertTrue(diagnoses.equals(expected_diagnoses))



"""Classify body measurements (anthropometrics) into different levels."""

import pandas as pd

from typing import LiteralString


def _classify_bmi_who2008(x):
    if pd.isna(x):
        return x
    if x < 18.5:
        return "underweight"
    if 18.5 <= x < 25.0:
        return "normal"
    elif 25.0 <= x < 30.0:
        return "overweight"
    elif 30.0 <= x < 35.0:
        return "obesity-CI"
    elif 35.0 <= x < 40.0:
        return "obesity-CII"
    elif x >= 40.0:
        return "obesity-CIII"
    raise ValueError(f"Unknown BMI value x={x}.")


def classify_body_mass_index(
    values: pd.Series, criterion: LiteralString = "who-2008"
) -> pd.Series:
    """Classify body mass index according to World Health Organization cut-off points.

    Classes:
      - `underweight`: <18.5
      - `normal`: 18.5–24.9
      - `overweight`: 25.0–29.9
      - `obesity-CI`: 30.0–34.9
      - `obesity-CII`: 35.0–39.9
      - `obesity-CIII`: ≥40.0

    Args:
        values: Body mass index (kg/m²).
        criterion: Use cut-off points from this organization (`"who-2008"`: Ref. [1]).
    Returns:
        A pandas Series with the corresponding discretized levels.

    References:
        [1] WHO Expert Consultation. Waist circumference and waist-hip ratio. Report of
        a WHO Expert Consultation. Geneva: World Health Organization, 2008:8–11 (2008).
    """
    result: pd.Series = values.map(_classify_bmi_who2008)
    dtype = pd.CategoricalDtype(
        categories=[
            "underweight",
            "normal",
            "overweight",
            "obesity-CI",
            "obesity-CII",
            "obesity-CIII",
        ],
        ordered=True,
    )
    return result.astype(dtype)



"""Classify various measurements pertaining to blood (e.g., pressure)."""

from typing import LiteralString

from numpy import where
import pandas as pd


def _classify_pressure_ehj2024(systolic, diastolic) -> pd.Series:
    blood_pressure_levels = pd.Series(index=systolic.index, dtype=str)
    is_normal = (systolic < 120) & (diastolic < 70)
    blood_pressure_levels[is_normal] = "normal"
    is_elevated = (
        ((systolic >= 120) | (diastolic >= 70)) & (diastolic < 90) & (systolic < 140)
    )
    blood_pressure_levels[is_elevated] = "elevated"
    is_hypertension = (systolic >= 140) | (diastolic >= 90)
    blood_pressure_levels[is_hypertension] = "hypertension"

    # All levels not discretized must be missing.
    is_any = is_normal | is_elevated | is_hypertension
    assert blood_pressure_levels[~is_any].isna().all(), "The levels are non-exhaustive!"
    dtype = pd.CategoricalDtype(
        categories=["normal", "elevated", "hypertension"], ordered=True
    )
    return blood_pressure_levels.astype(dtype)


def classify_blood_pressure(
    systolic: pd.Series, diastolic: pd.Series, criterion: LiteralString = "ESC-2024"
) -> pd.Series:
    """Classify blood pressure according to European Society of Cardiology guidelines.

    Classes:
      - `normal`: <120/70 mmHg
      - `elevated`: 120–139/70–89 mmHg
      - `hypertension`: ≥140/90

    Args:
        systolic: Systolic blood pressure (SBP) in mm Hg.
        diastolic: Diastolic blood pressure (DBP) in mm Hg.
        criterion: Use cut-off points from this guideline (`"ESC-2024"`: Ref. [1])
    Returns:
        A pandas Series with the corresponding discretized levels.

    References:
        [1] McEvoy et al., 2024 ESC Guidelines for the management of elevated blood pressure
        and hypertension  European Heart Journal 45, p. 3936 (2024): ehae178.
    """
    if not all(systolic.index == diastolic.index):
        raise KeyError("The systolic and diastolic indices do not match.")

    return _classify_pressure_ehj2024(systolic, diastolic)


def _has_elevated_waist_circumference(
    waist: pd.Series, sex: pd.Series, ethnicity: pd.Series | None = None
) -> pd.Series:
    """Whether the waist circumference is elevated according to NCEP ATP III 2005.

    Elevated waist circumference:
    Men: ≥1.02 m (≥0.90 m for Asian ancestry),
    Women: ≥0.88 m (≥0.80 m for Asian ancestry).

    Args:
        waist: circumference (m).
    """
    waist_score = pd.Series(index=waist.index)

    # When ethnicity is missing, assume not Asian.
    if ethnicity is None:
        ethnicity = pd.Series("white", index=waist_score.index)
    ethnicity = ethnicity.fillna("white")

    is_asian = ethnicity == "asian"
    is_male = sex == "male"
    is_female = sex == "female"
    big_waist = (
        ~is_asian & ((is_male & (waist >= 1.02)) | (is_female & (waist >= 0.88)))
    ) | (is_asian & ((is_male & (waist >= 0.90)) | (is_female & (waist >= 0.80))))
    waist_score.loc[big_waist] = 1
    normal_waist = (
        ~is_asian & ((is_male & (waist < 1.02)) | (is_female & (waist < 0.88)))
    ) | (is_asian & ((is_male & (waist < 0.90)) | (is_female & (waist < 0.80))))
    waist_score.loc[normal_waist] = 0
    return waist_score


def _has_elevated_triglycerides(
    triglcycerides, use_triglyceride_rx: pd.Series | None = None
):
    """Elevated triglycerides: ≥1.96 mmol/l or triglyceride lowering prescription."""
    trig_score = pd.Series(index=triglcycerides.index)

    # Assume no triglyceride lowering medication usage when the values are absent.
    if use_triglyceride_rx is None:
        use_triglyceride_rx = pd.Series(False, index=trig_score.index)
    use_triglyceride_rx = use_triglyceride_rx.fillna(False)

    trig_score.loc[(triglcycerides >= 1.96) | use_triglyceride_rx] = 1
    trig_score.loc[(triglcycerides < 1.96) & ~use_triglyceride_rx] = 0
    return trig_score


def _has_reduced_hdl_c(
    hdl: pd.Series, sex: pd.Series, use_hdl_rx: pd.Series | None = None
) -> pd.Series:
    """If the HDL cholesterol level are reduced according to NCEP ATP III 2005.

    - HDL prescription
    - Men: <1.0 mmol/l (40 mg/dL)
    - Women: <1.3 mmol/l (50 mg/dL)"""
    hdl_score = pd.Series(index=hdl.index)

    # Assume no HDL cholesterol low medication usage when the values are absent.
    if use_hdl_rx is None:
        use_hdl_rx = pd.Series(False, index=hdl_score.index)
    use_hdl_rx = use_hdl_rx.fillna(False)

    is_male = sex == "male"
    is_female = sex == "female"
    # N.B., Because of missing data, `is_low_hdl` is not the complement of
    # `is_normal_hdl`.
    is_low_hdl = (is_male & (hdl < 1.0)) | (is_female & (hdl < 1.3))
    hdl_score.loc[is_low_hdl | use_hdl_rx] = 1
    is_normal_hdl = (is_male & (hdl >= 1.0)) | (is_female & (hdl >= 1.3))
    hdl_score.loc[is_normal_hdl & ~use_hdl_rx] = 0
    return hdl_score


def _has_elevated_blood_pressure(
    systolic: pd.Series,
    diastolic: pd.Series,
    use_antihypertensive_meds: pd.Series | None = None,
) -> pd.Series:
    """Whether the blood pressure is elevated according to NCEP ATP III 2005.

    Elevated blood pressure:
    - systolic blood pressure ≥130 mm Hg
    - or diastolic blood pressure ≥85 mmHg
    - and/or use of antihypertensive medications.
    """
    bp_score = pd.Series(index=systolic.index)

    # Assume no antihypertensive medication usage when the values are absent.
    if use_antihypertensive_meds is None:
        use_antihypertensive_meds = pd.Series(False, index=bp_score.index)
    use_antihypertensive_meds = use_antihypertensive_meds.fillna(False)

    is_hypertension = (systolic >= 130) | (diastolic >= 85)
    bp_score.loc[is_hypertension | use_antihypertensive_meds] = 1

    is_normotension = (systolic < 130) & (diastolic < 85)
    bp_score.loc[~use_antihypertensive_meds & is_normotension] = 0
    return bp_score


def _has_elevated_fasting_glucose(
    glucose: pd.Series, use_glucose_rx: pd.Series | None = None
):
    """Elevated glucose: ≥5.6 mmol/l (100 mg/dl) or glucose lowering prescription."""
    glucose_score = pd.Series(index=glucose.index)

    # Assume no glucose lowering medication usage when the values are absent.
    if use_glucose_rx is None:
        use_glucose_rx = pd.Series(False, index=glucose_score.index)
    use_glucose_rx = use_glucose_rx.fillna(False)

    glucose_score.loc[(glucose >= 5.6) | use_glucose_rx] = 1
    glucose_score.loc[(glucose < 5.6) & ~use_glucose_rx] = 0
    return glucose_score


def _score_mets_ncep_atp_iii_2005(
    waist: pd.Series,
    triglycerides: pd.Series,
    hdl_cholesterol: pd.Series,
    systolic: pd.Series,
    diastolic: pd.Series,
    glucose: pd.Series,
    sex: pd.Series,
    ethnicity: pd.Series | None = None,
    use_triglyceride_meds: pd.Series | None = None,
    use_hdlc_meds: pd.Series | None = None,
    use_blood_pressure_meds: pd.Series | None = None,
    use_glucose_meds: pd.Series | None = None,
) -> pd.DataFrame:
    """Score the metabolic syndrome according to NCEP ATP III 2005."""
    score_card = {}

    # (1) Waist circumference
    score_card["waist"] = _has_elevated_waist_circumference(waist, sex, ethnicity)
    # (2) Triglycerides
    score_card["triglycerides"] = _has_elevated_triglycerides(
        triglycerides, use_triglyceride_meds
    )
    # (3) HDL cholesterol
    score_card["hdl_cholesterol"] = _has_reduced_hdl_c(
        hdl_cholesterol, sex, use_hdlc_meds
    )
    # (4) Blood pressure
    score_card["blood_pressure"] = _has_elevated_blood_pressure(
        systolic, diastolic, use_blood_pressure_meds
    )
    # (5) Fasting glucose
    score_card["glucose"] = _has_elevated_fasting_glucose(glucose, use_glucose_meds)

    return pd.concat(score_card, axis=1)


def diagnose_metabolic_syndrome(
    waist: pd.Series,
    triglycerides: pd.Series,
    hdl_cholesterol: pd.Series,
    systolic: pd.Series,
    diastolic: pd.Series,
    glucose: pd.Series,
    sex: pd.Series,
    ethnicity: pd.Series | None = None,
    use_triglyceride_meds: pd.Series | None = None,
    use_hdlc_meds: pd.Series | None = None,
    use_blood_pressure_meds: pd.Series | None = None,
    use_glucose_meds: pd.Series | None = None,
    criterion: LiteralString = "NCEP-ATP-III-2005",
) -> pd.Series:
    """Diagnose metabolic syndrome according to clinical guidelines.

    According to NCEP ATP III 2005 (`criterion="NCEP-ATP-III-2005"`), any three of the
    following five criteria:

      - Elevated waist circumference [men: ≥1.02 m (≥0.90 m for Asian ancestry),
          women: ≥0.88 m (≥0.80 m for Asian ancestry)].
      - Elevated triglycerides (≥1.96 mmol/l or triglyceride lowering medication)
      - Reduced HDL cholesterol (men: <1.0 mmol/l, women: <1.3 mmol/l, or HDL-C
          lowering medication).
      - Elevated blood pressure (systolic blood pressure ≥130 mm Hg, or diastolic
          blood pressure ≥85 mm Hg, and/or use of antihypertensive medications).
      - Elevated fasting glucose (≥5.6 mmol/l or glucose lowering medication).

    Args:
        waist: Circumference (m).
        triglycerides: Concentration (mmol/l).
        glucose: fasting glucose in (mmol/l).
        systolic: Systolic blood pressure, SBP (mm Hg).
        diastolic: Diastolic blood pressure, DBP (mm Hg).
        hdl_cholesterol: high-density lipoprotein cholesterol (mmol/l).
        sex: Either `"male"` or `"female"`.
        ethnicity: Use ethnicity specific waist circumferences. Values: `"asian"` or
            anything else.
        use_blood_pressure_meds: A boolean array indicating the use of antihypertensive
            medication. When absent, assume no prescription.
        use_hdlc_meds: A boolean array indicating the use of HDL-C lowering medication.
            When absent, assume no prescription.
        use_glucose_meds: A boolean array indicating the use of glucose lowering
            medication. When absent, assume no prescription.
        use_triglyceride_meds: A boolean array indicating the use of triglyceride
            lowering medication. When absent, assume no prescription.
        criterion: Use cut-off points according to this guideline (`"NCEP-ATP-III-2005"`:
            see Table 2, in Ref. [1]).

    Returns:
        Array indicating whether the patient has metabolic syndrome (1) or not (0).

    References:
        [1] Grundy, Scott M., et al. "Diagnosis and management of the metabolic
        syndrome: an American Heart Association/National Heart, Lung, and Blood
        Institute scientific statement." Circulation 112.17 (2005): 2735-2752.
    """
    score_card = _score_mets_ncep_atp_iii_2005(
        waist,
        triglycerides,
        hdl_cholesterol,
        systolic,
        diastolic,
        glucose,
        sex,
        ethnicity,
        use_triglyceride_meds,
        use_hdlc_meds,
        use_blood_pressure_meds,
        use_glucose_meds,
    )
    # For patients with missing values, we only know they are positive if they score
    # 3 or more.
    diagnoses = pd.Series(index=score_card.index, dtype="Int64")
    is_positive = score_card.sum(axis=1) >= 3
    diagnoses[is_positive] = 1
    is_negative = ~is_positive & score_card.notna().all(axis=1)
    diagnoses[is_negative] = 0
    return diagnoses



from unittest import TestCase

import pandas as pd

from clinbin.ehr import classify_chronic_conditions, _is_in_group


class TestIsInGroup(TestCase):
    def test_icd10_is_in_group(self):
        """Test if hierarchy of ICD10 codes is correctly handled."""
        group = {
            "E08.630",
            "E09.630",
            "E10.630",
            "E11.630",
            "E13.630",
            "K04.5",
            "K05",
            "K08.12",
            "K08.42",
            "C67",
        }

        test_codes = {
            "K05.0": True,  # Child of K05
            "E08.63": False,  # Too coarse
            "E08.631": False,  # Sibling of E08.630 (not a child)
            "K04.5": True,  # Exact Match
            "K04.51": True,  # Child of K04.5
            "Z99": False,  # Unrelated
            "K05": True,  # Exact match
            "K": False,  # Too coarse
            "E08.6": False,  # Too coarse
            "C67.9": True,
        }

        for code, expected in test_codes.items():
            with self.subTest(code=code, expected=expected):
                result = _is_in_group(code, set(group))
                self.assertEqual(result, expected)

    def test_classify_chronic_conditions(self):
        """Test the example in the docstring."""
        records = pd.DataFrame(
            {
                "ICD-10-CM": ["K05.0", "E08.63", "H35.2", "I11.0", "K04.51", "I25.7"],
                "ICD-10-PCS": [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, "B213"],
            }
        )
        chronic_conds = classify_chronic_conditions(records, criterion="wei-2024")
        # Record 1: child of K05.
        self.assertEqual(chronic_conds.loc[0, ("oral", "Periodontal disease")], 1)
        self.assertEqual(chronic_conds.loc[0].sum(), 1)  # Rest zero
        # Record 2: too coarse to fall within cataract (E08.36), but of diabetes (E08).
        self.assertEqual(chronic_conds.loc[1].sum(), 1)
        # Record 3: sibling of H35.3 (not a child).
        self.assertEqual(chronic_conds.loc[2].sum(), 0)
        # Record 4: I11.0 is in both "Congestive heart failure, Cardiomyopathy"
        # and "High blood pressure, hypertension".
        self.assertEqual(
            chronic_conds.loc[
                3, ("cardiovascular", "High blood pressure, hypertension")
            ],
            1,
        )
        self.assertEqual(
            chronic_conds.loc[
                3, ("cardiovascular", "Congestive heart failure, Cardiomyopathy")
            ],
            1,
        )
        self.assertEqual(chronic_conds.loc[3].sum(), 2)
        # Record 5: Child of K04.5.
        self.assertEqual(chronic_conds.loc[4, ("oral", "Periodontal disease")], 1)
        self.assertEqual(chronic_conds.loc[4].sum(), 1)  # Rest zero
        # Record 6: Has procedure code.
        self.assertEqual(
            chronic_conds.loc[
                5, ("cardiovascular", "Coronary artery bypass graft surgery (CABG)")
            ],
            1,  # B213
        )
        self.assertEqual(chronic_conds.loc[5, ("cardiovascular", "Angina")], 1)  # I25.7
        self.assertEqual(
            chronic_conds.loc[5, ("cardiovascular", "Coronary artery disease")], 1
        )  # I25
        self.assertEqual(chronic_conds.loc[5].sum(), 3)  # Rest zero



from unittest import TestCase

from clinbin import classify_body_mass_index
import numpy as np
import pandas as pd


class TestAnthropometry(TestCase):
    def test_body_mass_index(self):
        """Test the discretization of BMI."""
        bmi = pd.Series([18.49999, 18.5, np.nan, 26, 32, 38.123, 41.0])
        bmi_class = classify_body_mass_index(bmi)
        categories = pd.Categorical(
            [
                "underweight",
                "normal",
                np.nan,
                "overweight",
                "obesity-CI",
                "obesity-CII",
                "obesity-CIII",
            ],
            ordered=True,
            categories=[
                "underweight",
                "normal",
                "overweight",
                "obesity-CI",
                "obesity-CII",
                "obesity-CIII",
            ],
        )
        assert bmi_class.equals(pd.Series(categories, index=bmi.index))



"""Classify clinical variables using cut-off points from medical guidelines.

# Installation

```bash
pip3 install git+https://gitlab.com/hylkedonker/clinbin.git
```


"""

from .anthropometry import classify_body_mass_index
from .blood import classify_blood_pressure, diagnose_metabolic_syndrome
from .ehr import classify_chronic_conditions, get_wei2024_multimorbidity_weight_index



"""Ways to categorize electronic health record data."""

import json
from pathlib import Path
from typing import LiteralString

import pandas as pd


def _get_definitions_wei2024() -> tuple[
    dict[tuple[str, str], set],
    dict[tuple[str, str], set],
    dict[tuple[str, str], float],
]:
    """Lazy loading of data from JSON.

    Source: Supplementary Table 1 of BMJ open 14.2 (2024): e074390."""
    asset_dir = Path(__file__).parent / "assets"
    with open(asset_dir / "Wei_2024_multimorbidity_weighted_index.json", "r") as f:
        multimorbidity_weighted_index = json.load(f)
    # Make definition as flat dict with sets of ICD-10 codes.
    icd10cm_reference_definitions = {
        (kcond, kcat): payload["ICD-10-CM"]
        for kcond, v in multimorbidity_weighted_index.items()
        for kcat, payload in v.items()
    }

    icd10pcs_reference_definitions = {
        (kcond, kcat): payload["ICD-10-PCS"]
        for kcond, v in multimorbidity_weighted_index.items()
        for kcat, payload in v.items()
    }

    weights = {
        (kcond, kcat): payload["weight"]
        for kcond, v in multimorbidity_weighted_index.items()
        for kcat, payload in v.items()
    }
    return icd10cm_reference_definitions, icd10pcs_reference_definitions, weights


def _is_in_group(target: str, icd10_categories: set[str]) -> bool:
    """Is target ICD10 code contained in or is a child of any code in `icd10_categories`."""
    for reference in icd10_categories:
        # 1. Exact match (e.g., K05 matches K05)
        if target == reference:
            return True

        # 2. Hierarchical match (e.g., K05 matches K05.0)
        # We append a dot to ensure we match the hierarchy boundary
        if target.startswith(reference):
            return True

    return False


def _get_chronic_conditions_matching(
    code: str, reference_definitions: dict[tuple[str, str], set]
) -> list[tuple[str, str]]:
    """Go through all conditions, and see if code matches any of the ICD10 codes."""
    hits = []
    for key, icd10_code_set in reference_definitions.items():
        if _is_in_group(code, icd10_code_set):
            hits.append(key)
    return hits


def classify_chronic_conditions(
    icd10_records: pd.DataFrame, criterion: LiteralString = "wei-2024"
) -> pd.DataFrame:
    r"""Mark the presence of chronic conditions based on diagnoses and procedure codes.

    N.B. The mapping is not isomorphic: one ICD-10 code can cover multiple conditions.

    Args:
        icd10_records: A dataframe with diagnosis codes from the 10th revision of the
            International Classification of Diseases (ICD-10)  (column name:
            `"ICD-10-CM"`) and optionally ICD-10 procedure codes (column name:
            `"ICD-10-PCS"`).
        criterion: Use chronic condition criteria based on this reference
            (`"wei-2024"`: chronic conditions defined in the  multimorbidity-weighted
            index of Wei et al. [1]).

    Returns:
        A dataframe of the same length as `icd10_records`; chronic conditions are marked
        along the columns, where a 0/1 indicates the absence/presence.

    Example:
        ```python
        from clinbin.ehr import classify_chronic_conditions
        import pandas as pd

        records = pd.DataFrame({
            'ICD-10-CM': ['K05.0', 'E08.63', 'E08.631', 'I11.0', 'K04.51', 'I25.7'],
            'ICD-10-PCS': [pd.NA, pd.NA, pd.NA, pd.NA, pd.NA, 'B213'],
        })
        chronic_conds = classify_chronic_conditions(records, criterion="wei-2024")
        ```

    References:
        [1] Wei, Melissa Y., et al. "Development and validation of new
        multimorbidity-weighted index for ICD-10-coded electronic health record
        and claims data: an observational study." BMJ open 14.2 (2024): e074390.
    """
    if "ICD-10-CM" not in icd10_records.columns:
        raise KeyError("No ICD-10-CM column found.")

    if criterion != "wei-2024":
        raise ValueError(
            f"Unknown criterion: {criterion}. Only 'wei-2024' is supported."
        )

    icd10cm_ref_definition, icd10pcs_ref_definition, _ = _get_definitions_wei2024()

    column_names = pd.MultiIndex.from_tuples(
        icd10cm_ref_definition.keys(), names=["condition", "category"]
    )

    output = pd.DataFrame(0, index=icd10_records.index, columns=column_names)

    # Process unique codes instead of codes row-by-row (= faster).
    # Why? The number of unique codes less or equal to the number of rows.
    for code in icd10_records["ICD-10-CM"].dropna().unique():
        # Select all row with this code.
        has_icd10_code = icd10_records["ICD-10-CM"] == code
        matching_conditions = _get_chronic_conditions_matching(
            code, icd10cm_ref_definition
        )
        output.loc[has_icd10_code, matching_conditions] = 1

    if "ICD-10-PCS" in icd10_records.columns:
        for code in icd10_records["ICD-10-PCS"].dropna().unique():
            has_procedure = icd10_records["ICD-10-PCS"] == code
            matching_conditions = _get_chronic_conditions_matching(
                code, icd10pcs_ref_definition
            )
            output.loc[has_procedure, matching_conditions] = 1

    return output.astype("category")


def get_wei2024_multimorbidity_weight_index() -> pd.Series:
    """Get physical functioning of chronic conditions (multimorbidity-weighted index).

    The weights are based on the multimorbidity-weighted index of Wei et al. [1].

    Returns:
        Per chronic condition, the weight of the condition.

    References:
        [1] Wei, Melissa Y., et al. "Development and validation of new
        multimorbidity-weighted index for ICD-10-coded electronic health record
        and claims data: an observational study." BMJ open 14.2 (2024): e074390.
    """
    _, _, weights = _get_definitions_wei2024()
    index = pd.MultiIndex.from_tuples(weights.keys(), names=["condition", "category"])
    return pd.Series(weights, index=index)
